/* Program to generate Harnian weather.
 * Liam Quin, 1991
 *
 * $Id: Harvest.c,v 1.15 91/03/17 10:32:38 lee Dist1 $
 *
 */

#include <stdio.h>

#ifdef BSD
# define strrchr rindex
# include <strings.h> /* may need strings.h instead */
#else
# include <string.h> /* may need strings.h instead */
# define srandom srand
# define random rand
#endif

#ifdef __STDC__
# define PROTO(type, name, args) type name args 
# define Assert(e) MyAssert(e, __FILE__, __LINE__, #e)
#else
# define PROTO(type, name, args) type name()
# define Assert(e) MyAssert(e, __FILE__, __LINE__, "e")
#endif

/** exit stati: **/
#define INTERNAL_ERROR 1
#define BAD_USAGE 2
#define ASSERT_FAILED 3

/** external functions... **/
extern long random();
extern void srandom();
extern char *getenv();

/** Useful macros **/
#define Printf (void) printf
#define Fprintf (void) fprintf
#define NearestFarthing(dbl) ((double) ((int) ((dbl) * 4)) / 4.0) /*CHECKME*/

/** internal functions that are used before they are defined: **/
PROTO(void, PrintAcre, (char *Purpose, double Value, double Total));
PROTO(double, String2Float, (char *String, char *Description));
PROTO(void, DetermineLandAreas, ());
PROTO(void, MakeFreeMen, ());
PROTO(void, MakeSerfs, ());
PROTO(double, MakeRevenues, ());
PROTO(double, HouseHoldBudget, ());
PROTO(double, AddRevenue, ());
PROTO(double, OutGoings, ());

/** Harn stuff, global for error reporting only **/
static double WeatherOffset = 0.0;
static double Acres = -1.0; /* total acres */
static double TenantAcres = -1.0;
static double CropAcres = -1.0;
static double PastureAcres = -1.0;
static double WoodsAcres = -1.0;
static double WasteAcres = -1.0;
static int SerfCount = 130; /* serf population */
static int SerfLabour = 0, SerfKind = 0, SerfAcres = 0;
static int FreeMenCount = 1; /* free population */
static int FreeMenRent = 1;
static int FreeMenAcres = 0;

static double WeatherIndex = 0.0; /* to be calculated... */
static double FiefIndex = 1.0; /* default */
static double LandQuality = 1.00; /* default */

/** This is what we are trying to compute: **/
static double Income = 0.0;

/** generic program variables (not too many of these, please Liam!) **/
char *progname = "progname is set by main()";
char *cmdname = NULL;
int Verbose = 0;

/** **/


void
PrintCommandAndProgNames()
{
    if (cmdname) Fprintf(stderr, "%s: ", cmdname);
    Fprintf(stderr, "%s: ", progname);
}

#define UsagePrintf PrintCommandAndProgNames(); Fprintf

void
Usage()
{
    UsagePrintf(stderr, "estimate a Harn village's harvest yield\n");
    UsagePrintf(stderr, "options are [-Vvdxoautcpw]\n");
    if (!Verbose) {
	UsagePrintf(stderr, "\tUse -xv for details of options.\n");
	return;
    }
    UsagePrintf(stderr, "Options are:\n");
    UsagePrintf(stderr, " -V     -- print Version information\n");
    UsagePrintf(stderr, " -v     -- be verbose\n");
    UsagePrintf(stderr, " -d n   -- set debug level to n [now %d]\n", Verbose);
    {
	char *ThisOne = "print this explanation";
	char *ShorterOne = "print a shorter explanation";
	char *LongerOne = "print a Longer explanation";

	if (Verbose == 1) {
	    UsagePrintf(stderr, " -xv    -- %s\n", ShorterOne);
	    UsagePrintf(stderr, " -xvv   -- %s\n", LongerOne);
	} else if (Verbose == 2) {
	    UsagePrintf(stderr, " -xv    -- %s\n", ShorterOne);
	    UsagePrintf(stderr, " -xvv   -- %s\n", ThisOne);
	    UsagePrintf(stderr, " -xvvv  -- print a book about Harn\n");
	} else {
	    UsagePrintf(stderr, " -xv    -- %s\n", ShorterOne);
	    UsagePrintf(stderr, " -xvv   -- %s...\n", ThisOne);
	    UsagePrintf(stderr, " -xvvv  -- (I lied about the book, see)\n");
	}
    }
    UsagePrintf(stderr, " -o 3.4 -- give a bias to the Weather Index\n");
    if (Verbose > 1) {
	UsagePrintf(stderr, "           This is a percentage; higher values\n");
	UsagePrintf(stderr, "           give better weather.\n");
	if (Verbose > 2) {
	    UsagePrintf(stderr, "           Range should be -100 ... +100\n");
	    /* I never said the extra information would be useful! */
	}
    }
    UsagePrintf(stderr, " -a n   -- Allocate n Gross Acres of land\n");
    if (Verbose > 1) {
	UsagePrintf(stderr, "        n can be fractional, includes wasteland\n");
    }
    UsagePrintf(stderr, " -u n   -- Allocate n acres of unusable wasteland\n");
    UsagePrintf(stderr, " -t n   -- Allocate n acres to tenant farmers\n");
    UsagePrintf(stderr, " -c n   -- Allocate n acres for croplands\n");
    UsagePrintf(stderr, " -p n   -- Allocate n acres for pasture\n");
    UsagePrintf(stderr, " -w n   -- Allocate n acres for woodlands\n");
    UsagePrintf(stderr, " -q n   -- Land Quality is n [default %0.2f]\n",
							LandQuality);
    if (Verbose > 1) {
	UsagePrintf(stderr, "           LQ usually ranges from 0.85 to 1.15\n");
	UsagePrintf(stderr, "\n");
	UsagePrintf(stderr, "Give Gross Acres or all of the others.\n");
	if (Verbose > 2) {
	    UsagePrintf(stderr, "Waste Acres is 3d6%% of Gross if omitted.\n");
	}
    }
    UsagePrintf(stderr, " -s n   -- There are n serfs in the village\n");
    if (Verbose > 1) {
	UsagePrintf(stderr, "           These are your farmers...\n");
    }
    UsagePrintf(stderr, " -f n   -- There are n free men in the village\n");
    if (Verbose > 1) {
	UsagePrintf(stderr, "           These are millers, salters,...\n");
    }
    return;
}

void
MyAssert(Expression, File, Line, What)
    int Expression;
    char *File;
    int Line;
    char *What;
{

#define NotNull(s, Empty, Null) ((s) ? (*(s) ? s : Empty) : Null)

    if (!Expression) {
	UsagePrintf(stderr, "%s: line %d: Assertion Failed: %s\n",
		NotNull(File, "empty filename??", "NULL filename??"),
		Line,
		NotNull(What, "empty expression??", "NULL expression??")
	);
	exit(ASSERT_FAILED);
    }
}

int
Random(Min, Max)
    int Min, Max;
{
    /* Unix random number generators tend to be alternately odd and even,
     * so we have to divide by a prime number to get a plausible result.
     * We then ignore the high-order bits so that the mod function is fair.
     * Finally, rand % (Max - Min) gives us 0 ... (Max - Min - 1), so
     * we need rand % (Max - Min + 1), and then add Min to the result.
     *
     * Note: if you use this function a lot, put it in a header file
     * surrounded by #ifdef USE_INLINE, and change the declaration to
     * inline int Random().  Then include the header file where you use
     * the function.
     */
    return (((random() / 211) & 0777777) % (Max - Min + 1)) + Min;
}

int
AddDice(HowMany, Sides)
    int HowMany;
    int Sides;
{
    register int i;

    int Total = 0;

    for (i = 0; i < HowMany; i++) {
	Total += Random(1, Sides);
    }

    return Total;
}

#define d(n) Random(1, n) /* for d(100), etc */

typedef struct {
    int MaxRoll;
    char *Description;
    double WeatherIndex;
} t_WeatherTable;

t_WeatherTable WeatherTable[] = {
    {	 20,	"cold wet",	0.50	},
    {	 40,	"cool wet",	0.75	},
    {	 60,	"average",	1.00	},
    {	 80,	"warm dry",	1.25	},
    {	100,	"hot dry",	1.50	},
    {	 -1,	"Program Error",0.00	}
};

double
MakeWeatherIndex(Offset)
    double Offset; /* Percent bias (e.g. negative for worse weather) */
{
    register int i;
    double TotalSoFar = 0.0;
    t_WeatherTable *wp;


    for (i = 0; i < 5; i++) {
	int Value = d(100) + Offset;

	if (Value > 100) Value = 100;
	if (Value < 0) Value = 0;

	for (wp = WeatherTable; Value > wp->MaxRoll; wp++) {
	    if (wp->MaxRoll < 0) {
		UsagePrintf(stderr,
		    "%s: %d: MakeWeatherIndex(%f): no entry in %s for %%roll %d\n",
		     __FILE__, __LINE__, Offset, "Weather Table", Value
		);
		exit(1);
	    }
	}
	
	TotalSoFar += wp->WeatherIndex;
	if (Verbose > 1) {
	    Printf("Weather in growing month %d: %s [%2.2f]\n",
						i, wp->Description, wp->WeatherIndex);
	}

    }

    if (Verbose) {
	Printf("Average Weather Index: %0.2f\n", TotalSoFar/5.00);
    }

    return TotalSoFar/5.00;
}

main(argc, argv)
    int argc;
    char *argv[];
{
    /** getopt stuff **/
    extern char *optarg;
    extern int optind;
    int NeedUsage = 0;
    int c;

    if ((progname = strrchr(argv[0], '/')) != (char *) NULL) {
	++progname; /* step over the trailing '/' to get to the filename */
    } else {
	progname = argv[0];
    }
    cmdname = getenv("CMDNAME"); /* possibly null, of course */

    srandom(getpid() * 37 + 4);

#if 0
    /* for testing the dice rolling... */
    for (c = 0; c < 30; c++) {
	printf("%d\t%d\t%d\n", d(100), AddDice(3, 6), AddDice(3,6));
    }
#endif

    /** Option processing **/
    while ((c = getopt(argc, argv, "Vvd:xhs:f:o:a:u:t:c:p:w:q:")) != -1) {
	switch (c) {
	case 'V': /* Version */
	    UsagePrintf(stderr, "$Revision: 1.15 $\n");
	    exit(0);
	case 'v': /* verbose */
	    Verbose++; /* can be repeated... */
	    break;
	case 'd': /* debug (actually sets Verbose) */
	    Verbose += atoi(optarg);
	    break;
	case 'x': /* explanation */
	case 'h': /* help */
	    if (NeedUsage == 0) NeedUsage = 1; /* unchanged if -1 */
	    break;
	case 'o': /* weather offset as a percent, e.g. +4.5 */
	    WeatherOffset = String2Float(optarg, "-o: Weather Offset");
	    break;
	case 'a': /* acres (total) */
	    Acres = String2Float(optarg, "-a: Demesne Acres");
	    break;
	case 'u': /* unusable (waste) acres */
	    WasteAcres = String2Float(optarg, "-u: Waste Acres");
	    break;
	case 't': /* tenant acres */
	    TenantAcres = String2Float(optarg, "-t: Tenant Acres");
	    break;
	case 'c': /* cropland acres */
	    CropAcres = String2Float(optarg, "-c: Cropland Acres");
	    break;
	case 'p': /* pasture acres */
	    PastureAcres = String2Float(optarg, "-p: Pasture Acres");
	    break;
	case 'w': /* woods acres */
	    WoodsAcres = String2Float(optarg, "-w: Wooded Acres");
	    break;
	case 'q': /* land quality */
	    LandQuality = String2Float(optarg, "-q: land quality");
	    break;
	case 'f':
	    FreeMenCount = (int) String2Float(optarg, "-f: number of freemen");

	    /* Paranoia: Check it was a whole number! */
	    if (FreeMenCount != (int) String2Float(optarg, "Free Men")) {
		UsagePrintf(stderr,
			"-f %s: You must have a whole number of freemen!\n",
					optarg);
		UsagePrintf(stderr,
			"use the -xvv option for more explanation.\n");
		exit(BAD_USAGE);
	    }
	    break;
	case 's':
	    SerfCount = (int) String2Float(optarg, "-s: number of serfs");

	    /* Paranoia: Check it was a whole number! */
	    if (SerfCount != (int) String2Float(optarg, "SerfCount")) {
		UsagePrintf(stderr,
			    "-s %s: You must have a whole number of serfs!\n",
			    optarg);
		UsagePrintf(stderr,
			    "use the -xvv option for more explanation.\n");
		exit(BAD_USAGE);
	    }
	    break;
	case '?':
	    NeedUsage = (-1); /* beware of old C "=-" operator */
	    break;
	}
    }
    if (NeedUsage) {
	Usage();
	exit (NeedUsage < 0 ? BAD_USAGE : 0);
    }

    if (optind < argc) {
	UsagePrintf(stderr, "%s: warning: %d argument%s ignored (%s%s)\n",
			progname,
			argc - optind,
			(argc - optind == 1) ? "" : "s",
			argv[optind],
			(argc - optind == 1) ? "" : " onwards..."
	);
    }

     /** Ensure that the various acre values make sense **/
    DetermineLandAreas();

    /* Let's have some villagers... */
    MakeFreeMen();
    MakeSerfs();

    /** Determine the various indices... **/
    WeatherIndex = MakeWeatherIndex(WeatherOffset);

    /** Now calculate the Revenues... **/
    Income += MakeRevenues();

    /** Print the Gross Revenues, to the nearest farthing: **/
    Printf("Total Revenues: %12.2fd\n", NearestFarthing(Income));


    /** now the outgoings **/

    Income += OutGoings();

    Printf("Total Profit: %12.2fd\n", NearestFarthing(Income));

    /* Assume you're a knight, so you pay a 5% tax (Bailifs pay 30%) */
    Income = AddRevenue(Income * .05, "Tax to liege-lord", Income);

    Printf("To Storage: %-12.2f\n", Income);

    Printf("\n\tWeather was %1.2f: ", WeatherIndex);
    if (WeatherIndex >= 1.5) {
	Printf("A frollickingly good ");
    } else if (WeatherIndex > 1.2) {
	Printf("A good ");
    } else if (WeatherIndex > 0.8) {
	Printf("A slightly %s-average ",
				WeatherIndex >= 1.00 ? "above" : "below");
    } else if (WeatherIndex > 0.8) {
	Printf("A poor ");
    } else {
	Printf("A diabolical ");
    }
    Printf("harvest\n");
    if (Income <= SerfCount * 12) {
	Printf("\tbut some of your poorer serfs died of starvation\n");
	if (Verbose) {
	    switch(d(3)) {
	    case 1:
		Printf("\tbecause of poor planning... maybe?\n");
		break;
	    case 2:
		Printf("\tAh well -- they're only serfs.  Who cares?\n");
		break;
	    case 3:
		Printf("\tThey'll breed again in time.  Probably.\n");
		break;
	    case 4:
		Printf("\tThe slave market in Orball is good this year.\n");
		break;
	    }
	}
    }

    exit(0);
}

double
HouseHoldBudget()
{
    double Total = 0.0;

    Total = AddRevenue(-3000.0, "Personal expenses", Total);
    Total = AddRevenue(-1800.0, "War Horse (necessary)", Total);
    Total = AddRevenue(-2000.0, "Family", Total);
	/* Note: actually spouse is 1500 and kids are 1000, but
	 * I am simplifying and calling it family.
	 */
    Total = AddRevenue(-2000.0, "Bailiff", Total);
    Total = AddRevenue(-900.0, "Clothier", Total);
    Total = AddRevenue(-1200.0, "2 Harpers", Total);
    Total = AddRevenue(-900.0, "1 Hideworker", Total);

    Total = AddRevenue(-Total * .15, "Savings from having a clothier", Total);

    return Total;
}

double
AddRevenue(Amount, What, Total)
    double Amount;
    char *What;
    double Total;
{
    int Negative = (Amount < 0.0);

    if (Negative) Amount = -Amount;

    Amount = NearestFarthing(Amount);
    if (Verbose) {
	if (Negative) {
	    Printf("%40.40s: (%10.2f) %10.2f\n", What, Amount, Total);
	} else {
	    Printf("%40.40s:  %10.2f  %10.2f\n", What, Amount, Total);
	}
    }
    return Negative ? Total - Amount : Total + Amount;
}

double
OutGoings()
{
    double Total = 0.0;

    /** [1] Household Budget **/
    Total += HouseHoldBudget();

    /** [2] Maintenance **/
    Total = AddRevenue(-Acres * FiefIndex, "Maintenance", Total);

    /** [3] Demesne Acres (irrelevant) **/

    /** [4] Cropland planted for next year **/
    /* (we assume it to be the same as for this year) */
    Total = AddRevenue(-CropAcres * (15 + 25), "Planting crops", Total);

    /** [6] Unused (irrelevant) (there is no 5, it seems) **/

    /** [7] Free Exports (do this later) **/

    /** [11] Woods Cleared (1% growth per year) **/
    Total = AddRevenue(-WoodsAcres*110*.01, "Keeping the forest back", Total);

    /** [10] Labour -- we assume that it all works out OK **/

    return Total;

}

typedef struct {
    int MaxRoll;
    char *Description;
    int Sides; /* for acres */
    int Dice; /* for acres */
    int Flags;
} t_FreeMenGeneratingTable;

/* the flags: */
#define F_NoRent 01
    /* A Yeoman pays no rent */
#define F_BusinessRights 2
    /* e.g. a miller pays for the right to use a millstone (who wouldn't?) */
#define F_Miller 4
    /* they pay more rent */
#define F_LowRent 8
    /* token rent only */

t_FreeMenGeneratingTable FreeMenGeneratingTable[] = {
    {	 19,	"Miserable Miller     ",  6, 10, F_BusinessRights|F_Miller},
    {	 29,	"Jolly Miller         ",  6, 10, F_BusinessRights|F_Miller},
    {	 44,	"Yeoman               ",  3, 40, F_NoRent		},
    {	 50,	"Priest               ",  6,  5, F_LowRent		},
    {	 54,	"Priest with living   ",  6,  5, F_NoRent		},
    {	 64,	"Salter               ",  10, 1, F_BusinessRights	},
    {	 65,	"Jmorvi Metalsmith    ",  11, 1, F_BusinessRights	},
    {	 79,	"Metalsmith           ",  10, 1, F_BusinessRights	},
    {	 89,	"Woodcrafter          ",  10, 1, F_BusinessRights	},
    {	 97,	"Timberwright         ",  10, 1, F_BusinessRights	},
    {	100,	"Charcoaler           ",  10, 1, F_BusinessRights	},
    {	 -1,	"Malicious program bug",  -1, -1,	0	}
};

void
MakeFreeMen()
{
    t_FreeMenGeneratingTable *Tablep;
    register int i;

    for (i = 0; i < FreeMenCount; i++) {
	int Roll;
	int RentRoll;
	int RightsRoll;
	int AcresRoll;
	int Labour;

	Roll = d(100);

	for (Tablep = FreeMenGeneratingTable; Tablep->MaxRoll<Roll; Tablep++) {
	    if (Tablep->MaxRoll > 100) {
		UsagePrintf(stderr, "Internal Error: %s: %d: %d > 100\n",
				__FILE__, __LINE__ - 1, Tablep->MaxRoll);
		exit(INTERNAL_ERROR);
	    }
	}

	AcresRoll = AddDice(Tablep->Dice, Tablep->Sides);
	FreeMenAcres += AcresRoll;

	if (!(Tablep->Flags & F_NoRent)) {
	    if (Tablep->Flags & F_LowRent) {
		RentRoll = AcresRoll * LandQuality;
	    } else {
		int PerAcre = AddDice(2, 6);

		RentRoll =  PerAcre * AcresRoll * LandQuality;

		switch (PerAcre) {
		case 2:
		    Labour = 2 * AcresRoll;
		    break;
		case 3:
		    Labour = 1 * AcresRoll;
		    break;
		default:
		    Labour = 0;
		    break;
		}
	    }
	    FreeMenRent += RentRoll + Labour;
	} else {
	    RentRoll = 0;
	}

	if (Tablep->Flags & F_BusinessRights) {
	    if (Tablep->Flags & F_Miller) {
		RightsRoll = 240;
	    } else {
		RightsRoll = AddDice(10, 6);
	    }
	    FreeMenRent += RightsRoll;
	} else {
	    RightsRoll = 0;
	}

	if (Verbose > 3) {
	    Printf("%5d\t%s Pays %d rent (%d acre%s), %d rights, %d labour\n",
			i + 1, Tablep->Description, RentRoll, AcresRoll,
			AcresRoll == 1 ? "" : "s", RightsRoll, Labour
	    );
	}
    }
}

typedef struct {
    int MaxRoll;
    char *Description;
    int Acres;
    int Labour;
    int Kind;
} t_SerfGeneratingTable;

t_SerfGeneratingTable SerfGeneratingTable[] = {
    {	 1,	"Naked Slave          ",  0,	300,	 0	}, /* ouch! */
    {	12,	"Naked Cottar         ",  1,	 60,	 4	},
    {	19,	"Barefoot Cottar      ",  3,	 70,	 5	},
    {	29,	"Unhappy Cottar       ",  5,	 80,	 6	},
    {	36,	"Barefoot Half-Villein", 10,	100,	 7	},
    {	52,	"Half-Villein         ", 15,	120,	 8	},
    {	60,	"Aspiring Half-Villein", 20,	140,	 9	},
    {	70,	"Barefoot Villein     ", 25,	160,	10	},
    {	90,	"Contented Villein    ", 30,	180,	11	},
    {  100,	"hard-Working Villein ", 35,	200,	12	},
    {  999,	"Wanton program bug!  ", -1,	 -1,	99	},
};

void
MakeSerfs()
{
    t_SerfGeneratingTable *Tablep;
    register int i;
    int Roll, Roll2;

    for (i = 0; i < SerfCount; i++) {
	Roll = d(100);

	for (Tablep = SerfGeneratingTable; Tablep->MaxRoll < Roll; Tablep++) {
	    if (Tablep->MaxRoll > 100) {
		UsagePrintf(stderr, "Internal Error: %s: %d: %d > 100\n",
				__FILE__, __LINE__ - 1, Tablep->MaxRoll);
		exit(INTERNAL_ERROR);
	    }
	}

	SerfAcres += Tablep->Acres;
	Roll = d(15) - 8; /* -7 ... +7 inclusive */
	SerfLabour += Tablep->Labour + Roll;
	Roll2 = d(5) - 3; /* -2 ... +2 inclusive */
	SerfKind += Tablep->Kind + Roll2;

	if (Verbose > 3) {
	    Printf(
		"%5d\t%s farms %d acres, owes %2d days labour, %2d in kind\n",
		i + 1, Tablep->Description,
		Tablep->Acres, Tablep->Labour + Roll, Tablep->Kind + Roll2
	    );
	}
    }
    /* We'll modify the incomes by the FI and WI, though */
    SerfLabour *= FiefIndex * FiefIndex * WeatherIndex * .8;
    SerfKind *= FiefIndex * WeatherIndex * WeatherIndex;
    SerfKind *= (100 - AddDice(3, 10)) / 100.0;
    /* the final .8 or so reduction reflects the serfs making petitions
     * based on poor weather, deaths, children, or whatever.  
     */
}


double
MakeRevenues()
{
    double Total = 0.0;
    double Revenue;
    int n;

    /** [1] tenant acres **/
    Revenue = FreeMenRent;

    if (Verbose > 2) {
	Printf("\t%dd in rent and associated payments from %d freemen\n",
		FreeMenRent, FreeMenCount);
    }
    Total = AddRevenue(Revenue, "Income from freemen", Total);

    Revenue = SerfLabour + SerfKind;
    if (Verbose > 2) {
	Printf("\t%d days labour + %d in kind from %d busily farming serfs\n",
		SerfLabour, SerfKind, SerfCount);
    }
    Total = AddRevenue(Revenue, "Income from tenant farmers", Total);

    /** [2] crops **/
    Revenue = CropAcres * 60 * LandQuality * FiefIndex * WeatherIndex;
    if (Verbose > 2) {
	Printf("\t%4.2f x 60d x LQ [%0.2f] x FI [%0.2f] * WI [%0.2f]\n",
		CropAcres, LandQuality, FiefIndex, WeatherIndex);
    }
    Total = AddRevenue(Revenue, "Income from crops", Total);

    /** [3] Pasture **/
    Revenue = PastureAcres * 25 * LandQuality * FiefIndex;
    if (Verbose > 2) {
	Printf("\t%4.2f x 25d x LQ [%0.2f] x FI [%0.2f]\n",
		PastureAcres, LandQuality, FiefIndex);
    }
    Total = AddRevenue(Revenue, "Income from pasture", Total);
    Total = AddRevenue(Revenue * .1, "Hideworker improves pasture by", Total);

    /** [4] Woods **/
    Revenue = WoodsAcres * (n = AddDice(2, 4));
    if (Verbose > 2) {
	Printf("%4.2f x %dd\n",
		WoodsAcres, n);
    }
    Total = AddRevenue(Revenue, "Income from woods", Total);

    /** [5] Free Imports (assume none) **/

    if (Verbose > 1) {
	Printf("\tFree Imports from vassal demesnes could be:\n");
	Printf("\t\ttheir gross acres x 3.0 x LandQuality (held by you)\n");
	Printf("\t\ttheir gross acres x 0.5 x LandQuality (held by knight)\n");
	if (Verbose > 2) {
	    Printf("\tBut you will have to add that yourself...\n");
	}
    }

    /** [6] Gross Revenue **/

    /** [7] From Storage **/

    /* since we have no history, we invent a number.  The number suggested
     * is gross acres, but we modify it a little so it's less obvious.
     * We'll approximate a bell curve by rolling 100 50-sided dice...
     * It's done in 2 parts to avoid losing precision: since we can't
     * force evaluation order with parentheses, the compiler could do
     * Acres * AddDice() first and then the division, and we could lose
     * some farthings...
     */

    Revenue = AddDice(100, 50) / 5000;
    Revenue *= Acres;
    Total = AddRevenue(Revenue, "Goods from storage", Total);

    return Total;
}

void
DetermineLandAreas()
{

    /* Make sure the land amounts are set */
    if (Acres <= 0.0) {
	if (TenantAcres < 0.0 || CropAcres <= 0.0 ||
				PastureAcres <= 0.0 || WoodsAcres <= 0.0) {
	    UsagePrintf(stderr,
		"give either -a or -t, -c, -p & -w; use -xv for details\n");
	    exit(BAD_USAGE);
	}
	Acres = TenantAcres + CropAcres + PastureAcres + WoodsAcres;
	if (WasteAcres < 0.0) {
	    double R;

	    /*    Waste = 3d6% x GrossAcres
	     * hence
	     *    Waste = 3d6% x (NetAcres + Waste)
	     * Let R be 3d6%.  Then
	     *    Waste x (1 - R) = R x NetAcres
	     * and
	     *    Waste = R x NetAcres / (1 - R)
	     */

	    R = AddDice(3, 6) / 100.0;
	    WasteAcres = R * Acres / (1 - R);
	    Acres += WasteAcres;
	}
    } else {
	double Remainder = Acres;

	/* In this case we were given gross acres... */

	if (WasteAcres < 0) {
	    WasteAcres = Acres * AddDice(3, 6) / 100.0;
	}

	Remainder -= WasteAcres; /* subtracted up to 18% */
	Assert(Remainder > 0.0);

	if (WoodsAcres < 0) {
	    WoodsAcres = (AddDice(3, 6) + 10.0) * Acres / 100.0;
	}

	Remainder -= WoodsAcres; /* subtracted up to 18 + 28 = 46% */
	Assert(Remainder > 0.0);

	if (TenantAcres < 0) {
	    TenantAcres = 0.3 * Acres;
	    if (TenantAcres > Remainder) {
		TenantAcres = AddDice(7, 10) * Remainder / 100.0;
		/* 7...70% of the remainder */
	    }
	}
	Remainder -= TenantAcres;
	Assert(Remainder > 0.0);

	if (CropAcres < 0) {
	    CropAcres = 0.5 * Remainder;
	}
	Remainder -= CropAcres;

	Assert(Remainder >= 0.0);

	PastureAcres = Remainder;

	/* The rest is for pasture */

    }

    if (Verbose) {
	Printf("Total Acres: %g, of which %g are unusable;\n",
					Acres, WasteAcres);
	PrintAcre("crops", CropAcres, Acres);
	PrintAcre("tenant farming", TenantAcres, Acres);
	PrintAcre("pasture", PastureAcres, Acres);
	PrintAcre("woods", WoodsAcres, Acres);
	PrintAcre("waste", WasteAcres, Acres);
    }

    /** The land areas are now all determined. **/
}

void
PrintAcre(Purpose, Value, Total)
    char *Purpose;
    double Value;
    double Total;
{
    Printf("\t%4.3f acres of %s land [%0.2f%%]\n",
			Value, Purpose, Value * 100.0 / Total);
}


/** general useful functions **/

double
String2Float(String, Description)
    char *String;
    char *Description;
{
    /* We use strtod here.
     * If you don't have strtod, use atof() by compiling with -DUSE_ATOF
     */
#ifdef USE_ATOF
    PROTO(double, atof, (char *String));
#else
    PROTO(double, strtod, (char *String, char **EndPointer));
    double Result;
    char *EndOfNumber;
#endif

    if (!Description || !*Description) {
	if (!String || !*String) String = "NULL";

	UsagePrintf(stderr, "Internal Error: String2Float(%s, NULL)\n",
								String);
	exit(INTERNAL_ERROR);
    }
    if (!String || !*String) {
	if (!Description || !*Description) Description = "NULL";

	UsagePrintf(stderr, "Internal Error: String2Float(NULL, %s)\n",
								Description);
	exit(INTERNAL_ERROR);
    }
#ifdef USE_ATOF
    return atof(String);
#else
    Result = strtod(String, &EndOfNumber);
    if (!EndOfNumber) {
	UsagePrintf(stderr,
		"Internal Error: String2Float(%s, %s): strtod() failure\n",
		String, Description
	);
	exit(INTERNAL_ERROR);
    }
    if (EndOfNumber == String) {
	UsagePrintf(stderr, "%s: >>%s<< not understood as a number, sorry\n",
					    Description, String);
	UsagePrintf(stderr, "use the -x option for a usage message\n");
	exit(INTERNAL_ERROR);

    } else if (*EndOfNumber != '\0') {
	UsagePrintf(stderr,
		"%s: >>%s<< has unexpected trailing characters [%s]\n",
		Description, String, EndOfNumber
	);
	UsagePrintf(stderr, "use the -x option for a usage message\n");
	exit(INTERNAL_ERROR);
    }
    return Result;
#endif
}
/** **/
/* $Log:	Harvest.c,v $
 * Revision 1.15  91/03/17  10:32:38  lee
 * delinted.
 * 
 * Revision 1.14  91/03/17  10:24:27  lee
 * Seems to be OK, but a bit too random maybe?
 * 
 * Revision 1.13  91/03/17  09:51:47  lee
 * Well, it's all there, but it seems to random to me.
 * 
 * Revision 1.12  91/03/17  07:46:36  lee
 * it works.
 * 
 * Revision 1.11  91/03/17  07:40:39  lee
 * produce a sarcastic comment at the end.
 * 
 * Revision 1.10  91/03/17  07:18:16  lee
 * Now calls MakeSerf()... and started outgoingss, too.
 * 
 * Revision 1.9  91/03/17  06:45:21  lee
 * does household too now.
 * 
 * Revision 1.8  91/03/17  04:41:45  lee
 * Now at least attempts revenues...
 * 
 * Revision 1.7  91/03/17  03:50:15  lee
 * Compiles again, and does areas and weather...
 * 
 * Revision 1.6  91/03/17  03:37:22  lee
 * intermediate version.
 * 
 * Revision 1.5  91/03/17  03:00:02  lee
 * First plausible version.
 * 
 * Revision 1.4  91/03/17  02:49:54  lee
 * -xv and -xv are now eqivalent and work...
 * 
 * Revision 1.3  91/03/17  02:43:01  lee
 * More thorough in use of UsagePrint.
 * added a missing break; to the getopt switch().
 * 
 * Revision 1.2  91/03/17  02:36:23  lee
 * Added Usage() and made it more complicated...
 * 
 */
